// calc3.cc, C-compatible, numeric calculator.

// in reality the programm is a skeleton, for example, for an interactive
// statistical analysis programm.

// contents:

//   1. #include, global variables;

//   2. auxil. function (safe -- there is full error checking):

//         bool is_name(string name) -- check whether a command file 'name' was
//            opened already;

//         string get_bin(double n) -- return binary presentation of decimal
//            number 'n' in bash notation;

//         string get_in_base(double n, int base) -- return presentation of
//            a decimal number 'n' in base 'base';

//         struct blank, ostream& operator << -- print n spaces;

//         double usr_f1(double x) -- return factorial;

//         void clear(void) -- clear some global variables and add current
//            command to history list;

//        ifstream* open_file(const char* name) -- open a command file;

//        map<string,p_vector>::iterator
//         find(map<string,p_vector>& m, const string& str) -- find a
//             dynamically opened library by name;

//        void handler(int sig) -- signal handler;

//   3. command functions -- each function represents one command;

//   4. fEnum getFEnum() -- function to find commands in command line;

//   5. procOptions() -- function to process command line options.

//   6. main function.

// To add new command one must:
//     write a safe function for the command,
//     add element in fEnum enum,
//     add pointer to the function to f_arr array,
//     add branch into fEnum getFEnum() for the command.
//

/****************************************************************************
 1.
*/

#include <iostream>
#include <fstream>
#include <iomanip>
#include <strstream>
#include <string>
#include <stack>
#include <map>
#include <algorithm>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

extern "C" {
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
}

#include <forint/forint.hh>
#include <forint/config.h>

// maximum number of lines in history list file
const int CALC3_HISTORY_SIZE = 200;

// default number of lines to be saved by 'saveas' command
const int CALC3_DEFAULT_SAVEAS_SIZE = 100;

// default number of lines to be appended by 'appendto' command
const int CALC3_DEFAULT_APPENDTO_SIZE = 100;

// maximum length of a string containing a current working directory
const int CALC3_MAX_CWD_LENGTH = 300;

// maximum length of a string containing premessage
const int CALC3_EPREMESS_SIZE = 300;

using namespace FInt;

// global variables used in different functions...
string in_str;
string expr;
string first_token;
string second_token;
string third_token;
bool interactive = true;
char* inp = 0;
char* prev = 0;
int coun; // line counter
char* c_hist_name = 0;
bool exec_flag = false;
bool ESC_flag = true;
vector<string> file_name;
bool* user_break;
int locale;
char* help_text_p;
struct p_vector;
map<string,p_vector> lib_map;
// user interface object
mssg* c;
// forint object
forint* calc;
int base = 10;
string str_base = "dec";
bool no_parsing = false;
int k;
char* str_angle_unit = ".r";
map<string,double> ext_const;
map<string,double> ext_var;
unsigned int var_size = 0;
bool prog_flag  = false;
stack<istream*> input_stack;
istream* input  = static_cast<istream*>(&cin);
char epremess_buff[CALC3_EPREMESS_SIZE];
char wpremess_buff[CALC3_EPREMESS_SIZE];
char* epremess;
char* wpremess;
stack<long> line_number_stack;
string curr_file;
long curr_line;

enum actionEnum { SKIP_ACTION, BREAK_ACTION, CONTINUE_ACTION };

// this files content on-line help in English and Russian...
char* help_text =
#include "help.en"
;
char* rus_help_text =
#include "help.ru"
;

/****************************************************************************
 2.
*/

bool
is_name(string name)
{
  for (unsigned int i=0; i<file_name.size() && !*user_break; i++)
    if (name == file_name[i])
      return true;
  *user_break = false;
  return false;
}

// it's a fast function but up and down limits are too small and speed isn't
// crucial parameter here, so I use next function (get_in_base).

// for 64-bit `long int's
#undef LONG64

// make binary representation of a number correctly
string
get_bin(double n)
{
  if (!n || fabs(n) < 1) return "0";
  /* Number of decimal digits of precision in a double */      // float.h
  if (n >= pow(10,DBL_DIG+1)-1 ||
      n <= -pow(10,DBL_DIG+1)-1)
    return (locale)
? ":     "
:"Error: the number is out of range";
  long long tmp = (long long)n;
  long long i;
  string rv = (tmp > 0) ? "0" : "1";

#ifdef LONG64
  if (tmp >= 1<<31 || tmp <= -(1<<31)) i = 1<<63;
  else if (tmp >= 1<<15 || tmp <= -(1<<15)) i = 1<<31;

#else
  if (tmp >= 1<<15 || tmp <= -(1<<15)) i = 1<<31;
#endif

  else if (tmp >= 1<<7 || tmp <= -(1<<7)) i = 1<<15;
  else if (tmp >= 1<<3 || tmp <= -(1<<3)) i = 1<<7;
  else i = 1<<3;
  while (i>1 && !*user_break)
    rv += (tmp & (i>>=1)) ? "1" : "0";
  *user_break = false;
  return rv;
}

string
get_in_base(double n, int base)
{
  if (base > 36 || base < 2)
    return (locale)
      ? (!ESC_flag)
        ? ":      "
        : "\033[31m:      \033[00m"
      : (!ESC_flag)
        ? "Error: base is out of range"
        : "\033[31mError: base is out of range\033[00m";

  if ( !n || !((long long)n) ) return "0";

  //if (base == 2) return get_bin(n);
  /* Number of decimal digits of precision in a double */      // float.h
  if ( n >= pow(10,DBL_DIG+1)-1 ||
       n <= -pow(10,DBL_DIG+1)-1 )
    return (locale)
      ? (!ESC_flag)
        ? ":     "
        : "\033[31m:     \033[00m"
      : (!ESC_flag)
        ? "Error: the number is out of range"
        : "\033[31mError: the number is out of range\033[00m";

// in low case because '0' may be mixed up with 'O', 'Q'
  char sym [] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
		  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
		  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
		  'u', 'v', 'w', 'x', 'y', 'z' };

  string rv = "";
  char ch [3];
  ostrstream ostr (ch,3);
  if (base == 2) {
    if (n >= 0)
      c->mess('0',1,fg_red,bg_default,true);
    else
      c->mess('1',1,fg_blue,bg_default,true);
  }
  else if (base == 8) rv += "0";
  else if (base == 10) ;
  else if (base == 16) rv += "0x";
  else {
    ostr << base << ends;
    rv += ch + (string)"#";
  }

  long long ln = (long long)fabs(n);
  string str;
  int r=0;
  if ( base == 2 && n < 0 ) {
    char b_sym[2] = { '1', '0' };
    ln -= 1;
    while (ln && !*user_break) {
      r = ln % 2;
      ln /= 2;
      str += b_sym[r];
    }
  }
  else
    while (ln && !*user_break) {
      r = ln % base;
      ln /= base;
      str += sym[r];
    }
  *user_break = false;
  int i = str.size();
  while (i-- && !*user_break)
    rv += str[i];
  *user_break = false;
  return rv;
}

struct blank {
blank(ostream& out,unsigned int n) { while (n-- && !*user_break) out << ' '; }
};
ostream& operator << (ostream& out, const blank& bl) { return out; }

// external function
double
usr_f1(double x)
{
  if (x<0) {
if (locale)
  c->err(0,0,"   fact().\n");
else
  c->err(0,0,"fact(): negative argument.\n");
    c->flsh(); 
    return 0;
  }
  double y = 1;
  x = floor(x);
  while (x && !*user_break) {
    y *= x;
    --x;
  }
  *user_break = false;
  return y;
}

void
clear(void)
{
  if (interactive || exec_flag) {
    if (inp)
      if (prev && strcmp(inp,prev)) {
	free(prev);
	add_history (inp);
	prev = inp;
	++coun;
      }
      else
	free(inp);
    exec_flag = false;
  }
  //  in_str.replace(0,in_str.size(),1,' ');
  //  expr.replace(0,expr.size(),1,' ');
  in_str = "";
  expr   = "";
  c->flsh();
}

ifstream*
open_file(const char* name)
{
  ifstream* tmp = new ifstream(name);
  if (!tmp->good()) {
    if (locale)
      c->err(0,0,"   \'");
    
    else
      c->err(0,0,"failed to open file \'");
    c->err(name,5);
    c->err("\'\n");
    c->flsh(); 
    return 0;
  }
  return tmp;
}

enum ft_t { CLC3_V, CLC3_D, CLC3_DD } func_type;

struct p_vector {
  void* p;
  vector<string> v;
  vector<ft_t> t;
p_vector() { p=0;}
p_vector(void* l) { p=l;}
~p_vector() {}
p_vector (const p_vector& o) { p=o.p; v=o.v; t=o.t; }
p_vector&
operator = (const p_vector& o) {
  if (this != &o) {
    p=o.p;
    v=o.v;
    t=o.t;
  }
  return *this;
}
};

map<string,p_vector>::iterator
find(map<string,p_vector>& m, const string& str)
{
  map<string,p_vector>::iterator it;
  for (it=m.begin(); it != m.end() && !*user_break; it++)
    if (it->first == str) break;
  *user_break = false;
  return it;
}

void
handler(int sig)
{
  c->err(0,0,"");
  switch (sig) {
  case SIGINT:
if (locale)
  c->err("***  'q'   ***\n");
else
  c->err("*** Enter 'q' to quit ***\n");
    if (user_break)
      *user_break = true;
    if (signal(SIGINT,handler) == SIG_ERR) {
      c->err("*** calc3: signal(SIGINT) error ***\n");
    }
    break;
  case SIGFPE:
      c->err("*** Floating point exception ***\n");
    if (signal(SIGFPE,handler) == SIG_ERR)
      c->err("*** calc3: signal(SIGFPE) error ***\n");
    break;
  case SIGSEGV:
      c->err("*** Invalid memory reference ***\n");
    if (signal(SIGSEGV,handler) == SIG_ERR)
      c->err("*** calc3: signal(SIGSEGV) error ***\n");
    break;
  default:;
  }
  c->flsh(); 
}
/****************************************************************************
 3.
*/
actionEnum
f_quit()
{
  if (coun) {
    if (append_history (coun,c_hist_name)) {
      if(errno == ENOENT && write_history (c_hist_name))
	c->err(0,0,strerror(errno));
    }
    history_truncate_file(c_hist_name,CALC3_HISTORY_SIZE);
  }
  return BREAK_ACTION;
}
actionEnum
f_help()
{
// try to make help file it in /tmp and show by viewer,
// if failed just print help on display
  FILE* fd = 0;
  char tmp[100];
  bool is_printed = false;
  sprintf(tmp,"/tmp/calc3.txt_%d",getpid());
  if (!(fd=fopen(tmp,"w"))) {
    c->mess(0,help_text_p);
    //    c->mess("\n");
    is_printed = true;
  }
  else {
    fprintf(fd,help_text_p);
    fclose(fd);
  }
  if (!is_printed) {
    pid_t pid;
    int status;
    if ((pid = vfork()) < 0) {
      c->err(0,0,"calc3: vfork() error.\n");
      c->err(strerror(errno));
      c->err("\n");
    }
    else if (pid == 0)
      if (getenv("PAGER")) {
	if (execlp(getenv("PAGER"),getenv("PAGER"),tmp,0) == -1) {
	  c->err(0,0,"getenv(\"PAGER\") error");
	  c->err("\n");
	}
      }
      else if (execlp("less","less",tmp,0) == -1) {
	  c->err(0,0,"less error");
	  c->err("\n");
      }
      else if (execlp("more","more",tmp,0) == -1) {
	  c->err(0,0,"more error");
	  c->err("\n");
      }
      else {
	c->mess(0,help_text_p);
	//    c->mess("\n");
      }
    if (wait(&status) != pid) {
      c->err(0,0,"calc3: wait() error.\n");
      c->err(strerror(errno));
      c->err("\n");
    }
  }
  remove(tmp);
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_saveas()
{
  FILE* fd;
  if (!(fd=fopen(second_token.c_str(),"w"))) {
    c->err(0,0,strerror(errno));
    c->err("\n");
  }
  else {
    fclose(fd);
    char* e;
    int size = strtol(third_token.c_str(),&e,10);
    if ((size > 0 || (size=CALC3_DEFAULT_SAVEAS_SIZE)) &&
	append_history (size, second_token.c_str())) {
      c->err(0,0,strerror(errno));
      c->err("\n");
    }
  }
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_append_to()
{
  FILE* fd;
  if (!(fd=fopen(second_token.c_str(),"r")))
    if (errno == ENOENT) {
      fd=fopen(second_token.c_str(),"w");
      fclose(fd);
    } else {
      c->err(0,0,strerror(errno));
      c->err("\n");
    }
  else {
    fclose(fd);
    char* e;
    int size = strtol(third_token.c_str(),&e,10);
    if ((size > 0 || (size=CALC3_DEFAULT_APPENDTO_SIZE)) &&
	append_history (size, second_token.c_str())) {
      c->err(0,0,strerror(errno));
      c->err("\n");
    }
  }
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_load()
{
  if (first_token == "loadv") func_type=CLC3_V;
  else if (first_token == "loadd") func_type=CLC3_D;
  else if (first_token == "loaddd") func_type=CLC3_DD;
  if (!third_token.size()) {
if (locale)
  c->err(0,0,"  .\n");
else
  c->err(0,0,"function name has been missed.\n");
  }
  void* lib=0;
  void* pf;
  char *error;
  map<string,p_vector>::iterator lib_it = 0;
  if ((lib_it = find(lib_map,second_token)) !=  lib_map.end())
    lib = lib_map[second_token].p;
  else {
    lib = dlopen (second_token.c_str(), RTLD_LAZY);
    if (lib) {
      p_vector pv (lib);
      lib_map[second_token] = pv;
    }
  }
  if (!lib) {
    fputs (dlerror(), stderr);
    fputs ("\n", stderr);
  }
  else {
    switch (func_type) {
    case CLC3_V:
      pf = dlsym(lib,(third_token+(string)"__Fv").c_str());
      break;
    case CLC3_D:
      pf = dlsym(lib,(third_token+(string)"__Fd").c_str());
      break;
    case CLC3_DD:
      pf = dlsym(lib,(third_token+(string)"__Fdd").c_str());
      break;
    default:;
    }
    if ((error = dlerror()) != NULL)  {
      fputs(error, stderr);
      fputs ("\n", stderr);
    }
    else {
      int prv=0;
      switch (func_type) {
      case CLC3_V:
	prv=calc->putFunc0((double(*)(void))pf,
			  third_token.c_str(),true);
	break;
      case CLC3_D:
	prv=calc->putFunc1((double(*)(double))pf,
			  third_token.c_str(),true);
	break;
      case CLC3_DD:
	prv=calc->putFunc2D((double(*)(double,double))pf,
			   third_token.c_str(),true);
	break;
      default:;
      }
      if (!calc->error_code)
	lib_map[second_token].v.push_back(third_token);
      lib_map[second_token].t.push_back(func_type);
    }
  }
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_hex()
{
  if (first_token == "bin") {
    base=2;
    str_base = "bin";
  }
  else if (first_token == "oct") {
    base=8;
    str_base = "oct";
  }
  else if (first_token == "dec") {
    base=10;
    str_base = "dec";
  }
  else if (first_token == "hex") {
    base=16;
    str_base = "hex";
  }
  if (!second_token.size()) no_parsing = true;
  else expr.replace(0,3,1,' ');
  return SKIP_ACTION;
}
actionEnum
f_all()
{
  if (second_token.size())
    expr.replace(0,3,1,' ');
  else
    no_parsing = true;
  str_base = "all";
  return SKIP_ACTION;
}
actionEnum
f_base()
{
  char* e;
  int new_base = strtol(second_token.c_str(),&e,10);
  if (*e == '\0') {
    if (new_base >= 2 && new_base <= 36) {
      base = new_base;
      if (base == 2) str_base = "bin";
      else if (base == 8) str_base = "oct";
      else if (base == 10) str_base = "dec";
      else if (base == 16) str_base = "hex";
      else {
	char tmp [30];
	ostrstream ostr(tmp,30);
	ostr << base << ends;
	str_base = tmp;
      }
    }
    else {
if (locale) {
      c->err(0,0,"  .\n");
      c->warn("     [");
      c->warn("2");
      c->warn(",");
      c->warn("36");
      c->warn("].\n");
}
else {
      c->err(0,0,"Number's base.\n");
      c->warn("It must be in the interval [");
      c->warn("2");
      c->warn(",");
      c->warn("36");
      c->warn("].\n");
}
      c->flsh();
    }
    if (third_token.size())
      expr.replace(0,k,1,' ');
    else
      no_parsing = true;
  }
  return SKIP_ACTION;
}
actionEnum
f_angle()
{
  if (first_token == "rad" ) {
    calc->setAngleUnit(FI_RADIAN);
    str_angle_unit = ".r";
    if (!second_token.size())
      clear();
  }
  else if (first_token == "deg" ) {
    calc->setAngleUnit(FI_DEGREE);
    str_angle_unit = ".d";
    if (!second_token.size())
      clear();
  }
  else if (first_token == "grad" ) {
    calc->setAngleUnit(FI_GRAD);
    str_angle_unit = ".g";
    if (!second_token.size())
      clear();
  }
  return CONTINUE_ACTION;
}
actionEnum
f_putconst()
{
  char* e;
  if (ext_const.find(second_token) == ext_const.end() &&
      ext_var.find(second_token) == ext_var.end()) {
    ext_const[second_token] = strtod(third_token.c_str(),&e);
    if (!calc->putConst(ext_const[second_token],second_token.c_str()))
      var_size = (var_size < second_token.size())
	? second_token.size() : var_size;
  }
  else {
if (locale) {
  c->err(0,0," \'");
  c->err(second_token);
  c->err("\'  .\n");
}
else {
  c->err(0,0,"Name \'");
  c->err(second_token);
  c->err("\' is engaged already.\n");
}
  }
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_putvar()
{
  char* e;
  if (ext_var.find(second_token) == ext_var.end() &&
      ext_const.find(second_token) == ext_const.end()) {
    ext_var[second_token] = strtod(third_token.c_str(),&e);
    if (!calc->putVar(ext_var[second_token],second_token.c_str()))
      var_size = (var_size < second_token.size())
	? second_token.size() : var_size;
  }
  else {
if (locale) {
  c->err(0,0," \'");
  c->err(second_token);
  c->err("\'  .\n");
}
else {
  c->err(0,0,"Name \'");
  c->err(second_token);
  c->err("\' is engaged already.\n");
}
  }
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_showconst()
{
  for (map<string,double>::iterator it=ext_const.begin();
       it != ext_const.end() && !*user_break; it++) {
    c->mess(0,"  ");
    c->mess(it->first, fg_red, bg_default, true);
    c->mess(' ', var_size - it->first.size());
    c->mess(" = ");
    c->mess(it->second);
    c->mess("\n");
  }
  *user_break = false;
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_showvar()
{
  for (map<string,double>::iterator it=ext_var.begin();
       it != ext_var.end() && !*user_break; it++) {
    c->mess(0,"  ");
    c->mess(it->first, fg_green, bg_default, true);
    c->mess(' ', var_size - it->first.size());
    c->mess(" = ");
    c->mess(it->second);
    c->mess("\n");
  }
  *user_break = false;
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_showfunc()
{
  for (map<string,p_vector>::iterator it=lib_map.begin();
       it != lib_map.end() && !*user_break; it++) {
    c->mess(0,"\n");
    c->mess(it->first, fg_green, bg_default, true);
    c->mess(":\n");
    c->mess(' ', var_size - it->first.size());
    for (unsigned int k=0; k<it->second.v.size() && !*user_break; k++) {
      c->mess("   double ", fg_blue, bg_default, true);
      c->mess(it->second.v[k], fg_blue, bg_default, true);
      switch (it->second.t[k]) {
      case CLC3_V: c->mess(" (void)\n", fg_blue);
	break;
      case CLC3_D: c->mess(" (double)\n", fg_blue);
	break;
      case CLC3_DD: c->mess(" (double,double)\n", fg_blue);
	break;
      default:;
      }
    }
  }
  *user_break = false;
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_prog()
{
  prog_flag = !prog_flag;
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_exec()
{
  if (!second_token.size())
if (locale) {
  c->err(0,0,"     'exec'.\n");
  c->err("exec",5);
  c->err("'.\n");
}
else {
  c->err(0,0,"empty file name in the '");
  c->err("exec",5);
  c->err("' command.\n");
}
  else if (is_name(second_token) && !interactive)
if (locale) {
  c->err(0,0,"  '");
  c->err("exec",5);
  c->err(".\n");
}
else {
  c->err(0,0,"recursive '");
  c->err("exec",5);
  c->err(" command.\n");
}
  else {
    FILE* fd;
    input_stack.push(input);
    if (!(fd=fopen(second_token.c_str(),"r"))) {
      c->err(0,0,strerror(errno));
      c->err("\n");
    } else {
      fclose(fd);
      input = new ifstream(second_token.c_str());
      interactive = false;
      file_name.push_back(second_token);
      curr_file = second_token;
      line_number_stack.push(curr_line);
      curr_line = 0;
    }
  }
// because 'interactive = false' but this command 'exec ...' may be added
// to history list (if it doesn't coincide with the previous command)
  exec_flag = true;
  clear();
  return CONTINUE_ACTION;
}
actionEnum
f_shell()
{
  pid_t pid;
  int status;
  if ((pid = vfork()) < 0) {
      c->err(0,0,"calc3: vfork() error.\n");
      c->err(strerror(errno));
      c->err("\n");
  }
  else if (pid == 0)
    if (execlp(getenv("SHELL"),getenv("SHELL"),0) == -1) {
      c->err(0,0,"calc3: execlp() error.\n");
      c->err(strerror(errno));
      c->err("\n");
    }
  if (wait(&status) != pid) {
      c->err(0,0,"calc3: wait() error.\n");
      c->err(strerror(errno));
      c->err("\n");
  }
  clear();
  return CONTINUE_ACTION;
}

/****************************************************************************
 4.
*/
enum fEnum { FE_beg=0, FE_quit, FE_help, FE_saveas, FE_append_to,
	     FE_load, FE_hex, FE_all, FE_base, FE_angle,
	     FE_putconst, FE_putvar,
	     FE_showconst, FE_showvar, FE_showfunc,
	     FE_prog, FE_exec, FE_shell };

typedef actionEnum (*farr_t)();
farr_t f_arr [] = { 0, f_quit, f_help, f_saveas, f_append_to,
		    f_load, f_hex, f_all, f_base, f_angle,
		    f_putconst, f_putvar,
		    f_showconst, f_showvar, f_showfunc,
		    f_prog, f_exec, f_shell };
fEnum
getFEnum()
{
  fEnum rv = FE_beg;
// quit
  if (first_token == "q" || first_token == "quit")
    rv = FE_quit;
// print help
  else if (first_token == "h" || first_token == "help")
    rv = FE_help;
  else if (first_token == "saveas" && second_token.size())
    rv = FE_saveas;
  else if (first_token == "appendto" && second_token.size())
    rv = FE_append_to;
  else if (second_token.size() &&
	   (first_token == "loadv" ||
	    first_token == "loadd" ||
	    first_token == "loaddd"))
    rv = FE_load;
// change output base
  else if (first_token == "bin" ||
	   first_token == "oct" ||
	   first_token == "dec" ||
	   first_token == "hex")
    rv = FE_hex;
  else if (first_token == "all") {
    rv = FE_all;
  }
  else if (first_token == "base" && second_token.size())
    rv = FE_base;
// change angle unit
  else if (first_token == "rad" ||
	   first_token == "deg" ||
	   first_token == "grad")
    rv = FE_angle;
// put external constant
  else if (first_token == "putconst" && second_token.size())
    rv = FE_putconst;
// put external variable
  else if (first_token == "putvar" && second_token.size())
    rv = FE_putvar;
// show external constant
  else if (first_token == "showconst" && !second_token.size())
    rv = FE_showconst;
// show external variable
  else if (first_token == "showvar" && !second_token.size())
    rv = FE_showvar;
// show external functions
  else if (first_token == "showfunc" && !second_token.size())
    rv = FE_showvar;
// raise flag of current program's print of virtual machine's
  else if (first_token == "prog")
    rv = FE_prog;
// process command file(s)
// command file can exec ANOTHER command file
  else if (first_token == "exec")
    rv = FE_exec;
// run shell
  else if (first_token == "shell")
    rv = FE_shell;
  return rv;
}

/****************************************************************************
 5.
*/
void
procOptions(int argc, char** argv,
	    unsigned long& sl, mssg& a,
	    char* name, char* usage,
	    bool& batch, bool& calc_flag)
{
  for (int k=1; k<argc; k++) {
    ifstream* tmp_io = 0;
    if (strlen(argv[k]) > 2 &&
	!strncmp(argv[k],"--",2)) {
      if (!strcmp(&(argv[k][2]),"help")) {
	a.mess(0,name);
	a.mess("\n");
	a.mess(usage);
	a.mess("\n");
	a.flsh();
	exit(0);
      }
      else if (!strcmp(&(argv[k][2]),"exec") &&
	       argc > k + 1 &&
	       (tmp_io=open_file(argv[k+1]))) {
	k++;
	interactive = false;
	input_stack.push(input);
	file_name.push_back(argv[k]);
	curr_file = argv[k];
	line_number_stack.push(curr_line);
	curr_line = 0;
	input = tmp_io;
      }
      else if (!strcmp(&(argv[k][2]),"batch") &&
	       argc > k + 1 &&
	       (tmp_io=open_file(argv[k+1]))) {
	k++;
	interactive = false;
	input_stack.push(input);
	file_name.push_back(argv[k]);
	curr_file = argv[k];
	line_number_stack.push(curr_line);
	curr_line = 0;
	input = tmp_io;
	batch = true;
      }
      else if (!strcmp(&(argv[k][2]),"calc") &&
	       argc > k + 1) {
	interactive = false;
	calc_flag = true;
	in_str = inp = argv[k+1];
	k++;
      }
      else if (!strcmp(&(argv[k][2]),"noncolor")) {
	ESC_flag = false;
      }
      else if (!strcmp(&(argv[k][2]),"version")) {
if (locale) {
  a.mess(0,name);
  a.mess("\n");
  a.mess(": ");
  a.mess(FORINT_VERSION);
  a.mess("\n");
  a.mess("Copyright (c) 2000  \n");
} else {  
  a.mess(0,name);
  a.mess("\n");
  a.mess("Version: ");
  a.mess(FORINT_VERSION);
  a.mess("\n");
  a.mess("Copyright (c) 2000 Alexey Filin.\n");
}
	a.flsh();
	exit(0);
      }
      else {
	a.mess(0,usage);
	a.mess("\n");
	a.flsh();
	exit(0);
      }
    }
    else if (argv[k][0] == '-' && argv[k][1] != '\0' && argv[k][2] == '\0')
      switch (argv[k][1]) {
      case 'h':
	a.mess(0,name);
	a.mess("\n");
	a.mess(usage);
	a.mess("\n");
	a.flsh();
	exit(0);
	break;
      case 'e':
	if (argc < k + 2 ||
	    !(tmp_io=open_file(argv[k+1]))) {
	  a.mess(0,usage);
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	k++;
	interactive = false;
	input_stack.push(input);
	file_name.push_back(argv[k]);
	curr_file = argv[k];
	line_number_stack.push(curr_line);
	curr_line = 0;
	input = tmp_io;
	break;
      case 'b':
	if (argc < k + 2 ||
	    !(tmp_io=open_file(argv[k+1]))) {
	  a.mess(0,usage);
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	k++;
	interactive = false;
	input_stack.push(input);
	file_name.push_back(argv[k]);
	curr_file = argv[k];
	line_number_stack.push(curr_line);
	curr_line = 0;
	input = tmp_io;
	batch = true;
	break;
      case 'c':
	if (argc < k + 2) {
	  a.mess(0,usage);
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	interactive = false;
	calc_flag = true;
	in_str = inp = argv[k+1];
	k++;
	break;
      case 'n':
	ESC_flag = false;
	break;
      case 't':
	if (argc < k + 2) {
	  a.mess(0,usage);
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	char* end;
	sl = (unsigned int)abs(strtol(argv[k+1],&end,10));
	if (*end != '\0') {
	  a.mess(0,usage);
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	if (errno == ERANGE) {
	  a.mess(0,strerror(errno));
	  a.mess("\n");
	  a.flsh();
	  exit(0);
	}
	k++;
	break;
      default:
	a.mess(0,usage);
	a.mess("\n");
	a.flsh();
	exit(0);
      }
    else {
      a.mess(0,usage);
      a.mess("\n");
      a.flsh();
      exit(0);
    }
  }
}


/****************************************************************************
 6.
*/
main(int argc, char** argv)
{
  char* prompt = ".calc> ";
  double result;

  bool add_flag   = false;
  //bool prog_flag  = false;
  bool batch      = false;
  bool calc_flag  = false;
  bool exit_flag  = false;
  string prev_file = "";
  string hist_name  = string(getenv("HOME")) + string("/.history-calc3");
  c_hist_name = (char*)hist_name.c_str();
  fEnum fe = FE_beg;
  actionEnum rv;
  epremess = 0;
  wpremess = 0;
  prev = (char*)malloc(1);
  coun = 0; // line counter
  unsigned long sl=0;

  mssg a;
  locale = a.getLocale();
  help_text_p = (locale) ? rus_help_text : help_text;
  file_name.push_back(""); // dummy

  char* name = (locale)
    ? " c -    "
    "\033[05;31mc\033[05;33ma\033[05;32ml\033[05;36mc\033[05;34m3"
    "\033[00m"
    : "Numeric, C-like programmable calculator "
    "\033[05;31mc\033[05;33ma\033[05;32ml\033[05;36mc\033[05;34m3"
    "\033[00m";

  char* usage = (locale)
    ? ": calc3 [] [ ]\n\n"
    " ( ):\n"
    "  {-h, --help}       -      ,\n"
    "  {-e, --exec} FILE  -     ,\n"
    "  {-b, --batch} FILE -      ,\n"
    "  {-c, --calc} 'EXPRESSION' -        ,\n"
    "  --version          -      ,\n\n"
    " :\n"
    "  {-n, --noncolor}   -   (  ),\n"
    "  -t <>      - '' ,   .\n"
    : "Usage: calc3 [mode] [optional]\n\n"
    "mode (one of):\n"
    "  {-h, --help}       - print this help and exit,\n"
    "  {-e, --exec} FILE  - execute a command file before session,\n"
    "  {-b, --batch} FILE - execute a command file and exit,\n"
    "  {-c, --calc} 'EXPRESSION' - evaluate an expression and exit,\n"
    "  --version          - print version of calc3 and exit,\n\n"
    "optional:\n"
    "  {-n, --noncolor}   - non-colored output (for redirected output),\n"
    "  -t <delay>   - 'tipewriter' print, delay in millisecodns.\n";

  if (read_history (c_hist_name) && errno == ENOENT)
if (locale) {
  a.mess(0,"\n",fg_red);
  a.mess('', 1,fg_yellow);
  a.mess('', 1,fg_green);
  a.mess('', 1,fg_cyan);
  a.mess('', 1,fg_blue);
  a.mess('', 1,fg_magenta);
  a.mess(" ",fg_red);
  a.mess('', 1,fg_yellow);
  a.mess('', 1,fg_green);
  a.mess('', 1,fg_cyan);
  a.mess('', 1,fg_blue);
  a.mess('', 1,fg_magenta);
  a.mess('', 1,fg_red);
  a.mess('', 1,fg_yellow);
  a.mess('', 1,fg_green);
  a.mess('', 1,fg_cyan);
  a.mess('!', 1,fg_blue);
  a.mess('\n',2);
  a.mess(name);
  a.mess("\n");
  a.mess("Copyright (c) 2000  .\n"
	  " \'h\'    .\n\n");
  a.flsh();
} else {
  a.mess(0,"\n", fg_red);
  a.mess('W', 1,fg_yellow);
  a.mess('e', 1,fg_green);
  a.mess('l', 1,fg_cyan);
  a.mess('c', 1,fg_blue);
  a.mess('o', 1,fg_magenta);
  a.mess('m', 1,fg_red);
  a.mess('e', 1,fg_yellow);
  a.mess('!', 1,fg_green);
  a.mess('\n',2);
  a.mess(name);
  a.mess("\n");
  a.mess("Copyright (c) 2000 Alexey Filin.\n"
	  "Type \'h\' to see on-line help.\n\n");
  a.flsh();
}
  if (argc>1) {       
    procOptions(argc, argv,
		sl, a, name, usage,
		batch, calc_flag);
  }

// either colored print or not...
  if (!ESC_flag) c = new mssg;
  else c = new cmssg;
  /* Number of decimal digits of precision in a double */      // float.h
  c->precision(DBL_DIG);
  c->setSleep(sl);

// forint object
  calc = new forint (*c);

// put some constants
  calc->putConst(M_PI, "pi");
  calc->putConst(M_E, "e");

// put a function
  calc->putFunc1 (usr_f1, "fact");

// get pointer to break flag (to exit from an endless loop)
  user_break = calc->getPBrk();

  if (signal(SIGINT,SIG_IGN) != SIG_IGN)
    if (signal(SIGINT,handler) == SIG_ERR) {
      c->err(0,0,"calc3: signal(SIGINT) error.\n");
      c->flsh();
    }
  if (signal(SIGFPE,handler) == SIG_ERR) {
    c->err(0,0,"calc3: signal(SIGFPE) error.\n");
    c->flsh();
  }
  if (signal(SIGSEGV,handler) == SIG_ERR) {
    c->err(0,0,"calc3: signal(SIGSEGV) error.\n");
    c->flsh();
  }

// main loop
  while(1) {

// if it's interactive input type prompt
    if (interactive) {
      inp = readline((char*)(str_base + string(str_angle_unit) +
			     string(prompt)).c_str());
      if (inp) {
	char* ts = inp;
	while (ts && *ts==' ') ++ts;
	in_str = (ts)
	  ? string(ts)
	  : string("");
      }
      else {
	in_str = string("");
	c->mess("\n");
	c->flsh();
      }
    }
    else if (calc_flag) {
      calc_flag = false;
      exit_flag = true;
    }
    else if (exit_flag) exit(0);
// else check stream state and if EOF is occured remove stream and get
// previous one from the stack
    else if (input->eof()) {
      delete input;
      input = input_stack.top();
      input_stack.pop();
      file_name.pop_back(); // because on the top lies current opened file
      curr_file = file_name.back();
      curr_line = line_number_stack.top();
      line_number_stack.pop();
      interactive = (input == &cin) ? true : false;
      if (interactive) {
	epremess = 0;
	c->setEPremess(epremess);
	wpremess = 0;
	c->setWPremess(wpremess);
      }
      if (batch && input == &cin) exit(0);
      continue;
    }
    else {
      getline(*input,in_str);
      inp = (char*)in_str.c_str();
      ++curr_line;

      ostrstream ostr(epremess_buff,CALC3_EPREMESS_SIZE);
      ostr << curr_file << ':' << curr_line;
if (locale)
  ostr << ": : " << ends;
else
  ostr << ": Error: " << ends;
      epremess = epremess_buff;
      c->setEPremess(epremess);

      ostrstream ostr2(wpremess_buff,CALC3_EPREMESS_SIZE);
      ostr2 << curr_file << ':' << curr_line;
if (locale)
  ostr2 << ": : " << ends;
else
  ostr2 << ": Warning: " << ends;
      wpremess = wpremess_buff;
      c->setWPremess(wpremess);
    }

    if (in_str.size() && in_str[in_str.size()-1] == '\\') {
      if (interactive) {
	add_history (inp);
	if (prev) free(prev);
	prev = inp;
	++coun;
      }
      in_str = in_str.substr(0,in_str.size()-1);
      expr += in_str;
      prompt = ".add> ";
      add_flag = true;
      continue;
    }
    else if (add_flag) {
      if (interactive) {
	add_history (inp);
	if (prev) free(prev);
	prev = inp;
	++coun;
      }
      expr += in_str;
      prompt = ".calc> ";
      add_flag = false;
    }
    else
      expr = in_str;

// omit blanks
    int i=-1;
    int s_size = expr.size();
    while (s_size > 0 && expr[++i] == ' ') s_size--;
    if (!s_size) {
      clear();
      continue;
    }
// get token
    first_token  = "";
    i=0;
    s_size = expr.size();
    while (s_size-- > 0 && expr[i] != ' ')
      first_token += expr[i++];
    if (first_token.size() > 0
	&& first_token[0] == '/'
	&& first_token[1] == '/') {
      if (!(first_token.size() > 1 && first_token[2] == '/')){
	c->mess(0,expr.c_str(),fg_magenta,bg_default,true);
	c->mess("\n");
	c->flsh();
      }
      
      clear();
      continue;
    }

// omit blanks
    k = first_token.size();
    s_size = expr.size() - k;
    while (s_size && expr[++k] == ' ') s_size--;
// get second token if one
    second_token = "";
    s_size = expr.size() - k;
    while (s_size-- && expr[k] != ' ')
      second_token += expr[k++];

// [0,k] - substring consists of ONLY first and second tokens
// omit blanks
    int l = k;
    s_size = expr.size() - l;
    while (s_size-- && expr[++l] == ' ');
// get third token if one
    third_token  = "";
    s_size = expr.size() - l;
    while (s_size-- && expr[l] != ' ')
      third_token += expr[l++];

///////////////////////////
// find a command...
    if ((fe=getFEnum()))
// and execute it...
      rv = f_arr[fe]();
    else
      rv = SKIP_ACTION;
///////////////////////////

// make action after a found command...
    if (rv == BREAK_ACTION)
      break;
    else if (rv == CONTINUE_ACTION)
      continue;
    else if (rv == SKIP_ACTION)
      ;
    else {
      c->err(0,0,"Unknown command\n");
      c->flsh();
      break;
    }

// parse
    if (!no_parsing)
      if (calc->parse((char*)expr.c_str())) {
	clear();
	continue;
      }
// print VM programs
    if (prog_flag) {
      int length;
      char** h = calc->getRProg(length);
      c->mess(0,"RVM:",fg_green,bg_default,true);
      for (int i=0; i<length; i++) {
	c->mess(' ',1);
	c->mess(h[i],fg_cyan,bg_default,true);
      }
      c->mess("\n");
      h = calc->getSProg(length);
      c->mess("SVM:",fg_green,bg_default,true);
      for (int i=0; i<length; i++) {
	c->mess(' ',1);
	c->mess(h[i],fg_cyan,bg_default,true);
      }
      c->mess("\n");
      c->flsh();
    }
// if no errors occured calculate
    if (calc->error_code || !calc->getSProgSize()) {
      no_parsing = false;
      clear();
      continue;
    }
    else
      result = calc->calculate(SVM_MODE);
    if (calc->error_code) {
      no_parsing = false;
      clear();
      continue;
    }
    if (isinf(result)<0) {
if (locale)
  c->err(0,0,"***   ***\n");
else 
  c->err(0,0,"*** Negative  infinity ***\n");
      clear();
      continue;
    }
    if (isinf(result)>0) {
if (locale)
  c->err(0,0,"***   ***\n");
else 
  c->err(0,0,"*** Positive  infinity ***\n");
      clear();
      continue;
    }
    if (isnan(result)) {
if (locale)
  c->err(0,0,"***   ***\n");
else 
  c->err(0,0,"*** Not-a-number ***\n");
      clear();
      continue;
    }

// print the result
    if (str_base == "all") {
      if (result < 0)
if (locale)
  c->err(0,0," .\n");
else
  c->err(0,0,"negative number.\n");
      else if (result > LONG_MAX)
if (locale)
  c->err(0,0,".\n");
else
  c->err(0,0,"overflow.\n");
      else {
	c->mess(0,"      ");
	c->mess(get_in_base(result,10),fg_default,bg_default,true);
	c->mess("  ");
	c->mess(get_in_base(result,8),fg_default,bg_default,true);
	c->mess("  ");
	c->mess(get_in_base(result,16),fg_default,bg_default,true);
	c->mess("  \'");
	c->mess(char(result),1,fg_default,bg_default,true);
	c->mess("\'\n");
      }  
    }
    else if (base == 2) {
      c->mess(0,"      ");
      //c->mess(get_bin(result),fg_default,bg_default,true);
      c->mess(get_in_base(result,2),fg_default,bg_default,true);
      c->mess("\n");
    }
    else if (base == 10) {
      c->mess(0,"      ");
      c->mess(result,fg_default,bg_default,true);
      c->mess("\n");
    }
    else {
      c->mess(0,"      ");
      c->mess(get_in_base(result,base),fg_default,bg_default,true);
      c->mess("\n");
    }
    
// at last, reset variables
    clear();
    no_parsing = false;
    prompt = ".calc> ";
  }
}
